//
//  Layer.js
//  Kite
//
//  Created by Joe Rhodes
//  Copyright (c) 2015 Joe Rhodes. All rights reserved.
//


@import 'Helpers.js'
@import 'ShapePathWriter.js'
@import 'ParagraphStyle.js'

var sketch = require('sketch')


// -------------------------------------------------------------------------------------------
// Layer Classes
// -------------------------------------------------------------------------------------------

var kTextLayerClass               = "MSTextLayer"
var kSliceLayerClass              = "MSSliceLayer"
var kArtboardLayerClass           = "MSArtboardGroup"
var kShapeGroupLayerClass         = "MSShapeGroup"
var kLayerGroupLayerClass         = "MSLayerGroup"
var kBitmapLayerClass             = "MSBitmapLayer"
var kSymbolInstanceLayerClass     = "MSSymbolInstance"
var kSymbolMasterLayerClass       = "MSSymbolMaster"
var kShapePathLayerClass          = "MSShapePathLayer"


// -------------------------------------------------------------------------------------------
// Constants
// -------------------------------------------------------------------------------------------

var kNonZeroWindingRule          = 0
var kEvenOddWindingRule          = 1
var kDefaultShapeLayerMiterLimit = 10

var kCALayerLineJoinStyleMiter  = 0
var kCALayerLineJoinStyleRound  = 1
var kCALayerLineJoinStyleBevel  = 2

var kCALayerLineCapStyleButt    = 0
var kCALayerLineCapStyleRound   = 1
var kCALayerLineCapStyleSquare  = 2

var kCAFilterNearest = "nearest"
var kCAFilterLinear = "linear"
var kCAFilterTrilinear = "trilinear"

var kFillTypeSolidColor         = 0
var kFillTypeImage              = 4


// -------------------------------------------------------------------------------------------
// Layer
// -------------------------------------------------------------------------------------------

Layer = function(sketchLayer, exportInfo) {

    this.exportInfo = exportInfo
    this.className = this.layerClass
    this.sketchLayer = sketchLayer
    this.anchorPoint = { x: 0, y: 0 }
    this.uuid = [[NSUUID UUID] UUIDString]
    this.sublayers = []
    this.parentLayer = null
    this.duration = 3600
    this.name = sketchLayer.name()
    this.hidden = (sketchLayer.isVisible() == false)
    this.locked = sketchLayer.isLocked()
    this.contentsGravity = "center"
    this.contentsScale = exportInfo.exportOptions.imageExportScale
    this.masksToBounds = false
	this.minificationFilter = kCAFilterLinear
	this.magnificationFilter = kCAFilterLinear
    this.metadata = this.createMetadata()
}

Layer.prototype = Object.create(Object.prototype, {

    bounds: {
        get: function () {
            return {
                origin: {
                    x: 0,
                    y: 0
                },
                size: this.size
            }
        }
    },

    size: {
        get: function () {
            return {
                width: this.frame().width(),
                height: this.frame().height()
            }
        }
    },

    position: {
        get: function () {

            // See Position Fix below for details
            //
            if (this.needsPositionFix) {
                return this.fixedPosition
            }

            return {
                x: this.frame().x(),
                y: this.frame().y()
            }
        }
    },

    frame: {
        value: function() {

            if (this.exportsContentImage) {
                return this.influenceRect
            }
            else {
                return this.sketchLayer.frame()
            }
        }
    },

    influenceRect: {
        get: function() {
            return Helpers.influenceRectForSketchLayer(this.sketchLayer)
        }
    },

    opacity: {
        get: function() {

            if (this.hasStyle == false) {
                return 1.0
            }

            var style = this.sketchLayer.style()
            if (style == undefined) { return 1.0 }

            var contextSettings = style.contextSettings()
            if (contextSettings == null) { return 1.0 }

            return contextSettings.opacity()
        }
    },

    layerClass: {
        get: function() {
            return "CALayer"
        }
    },

    prepareForSave: {
        value: function() {

            // If we have a clipping mask, set the 'mask'
            // property of our parent to ourselves
            //
            if (this.sketchLayer.hasClippingMask())
            {
                this.parentLayer.mask = this
                this.masksParentLayer = true
            }

            if (this.exportsSublayers)
            {
                for (var sublayer of this.sublayers) {
                    sublayer.prepareForSave()
                }
            }
        }
    },

    save: {
        value: function(exportInfo) {
            log("Layer Save! (" + this.name + ")")

            if (this.exportsContentImage) {
                var contentIdentifier = exportInfo.assetManager.saveLayerAsset(this, exportInfo.sketchPage)

                if (contentIdentifier) {
                    this.contentIdentifier = contentIdentifier
                }
            }

            if (this.exportsSublayers) {
                for (var sublayer of this.sublayers) {
                    sublayer.save(exportInfo)
                }
            }
        }
    },

    hasSublayers: {
        get: function() {
            return false
        },
    },

    hasStyle: {
        get: function() {
            return (typeof this.sketchLayer.style === 'function');
        },
    },

    exportsContentImage: {
        get: function() {
            return true
        }
    },

    // If we are exporting an image of the layer, the shadow will be baked in the image,
    // so no need to add an additional native CALayer shadow
    //
    exportsShadow: {
        get: function() {
            return (this.exportInfo.exportOptions.exportShapeLayersAsImages == false)
        }
    },

    exportsSublayers: {
        get: function() {
            return true
        }
    },

    urlForImageOfLayerInDocument: {
        value: function(sketchDoc, sketchPage) {

            var slice = this.artboardOrSliceForLayer(sketchPage)
            var tempURL = this.temporaryImagePath

            [sketchDoc saveArtboardOrSlice:slice toFile:tempURL.path()]

            if (this.exportLayer) {
                this.exportLayer.removeFromParent()
                this.exportLayer = null
            }

            return tempURL
        }
    },

    // -------------------------------------------------------------------------------------------
    // Position Fix
    //
    // Workaround needed for layers not on the first page that are immediate sublayers of
    // the page. These layers report a bogus position value and it needs to be corrected by a
    // fixed offset value.
    // -------------------------------------------------------------------------------------------

    needsPositionFix: {
        get: function () {

            if ((this.parentLayer instanceof PageLayer) == false) { return false }
            var parentPageLayer = this.parentLayer

            if ((parentPageLayer.parentLayer instanceof ShapeContainerLayer) == false) { return false }
            var shapeContainerLayer = parentPageLayer.parentLayer

            if ((shapeContainerLayer.canvasInfo instanceof CanvasInfo) == false) { return false }
            var canvasInfo = shapeContainerLayer.canvasInfo

            // All layers that are a sublayer of a PageLayer where that PageLayer is not the the
            // first page layer need to have a fixed position value
            //
            return (canvasInfo.canvasID > 1)
        }
    },

    fixedPosition: {
        get: function() {

            var parentPageLayer = this.parentLayer
            var contentOrigin = parentPageLayer.contentBounds.origin

            return {
                x: this.frame().x() + Math.abs(contentOrigin.x),
                y: this.frame().y() + Math.abs(contentOrigin.y)
            }
        }
    },

    // -------------------------------------------------------------------------------------------
    // Helper
    // -------------------------------------------------------------------------------------------

    createMetadata: {
        value: function() {

            var sketchLayer = this.sketchLayer
            if (sketchLayer == null || sketchLayer == undefined || sketchLayer.objectID == undefined) { return null }

            var layerID = sketchLayer.objectID()
            if (layerID == null || layerID == undefined) { return null }

            var metadata = {
                className:  'JRMetadataDictionary',
                entries: [
                    {
                        value: {
                            type: 'String',
                            value: layerID
                        },
                        className: 'JRMetadataEntry',
                        key: 'sketchObjectID',
                        type: 3
                    }
                ]
            }

            return metadata
        }
    },

    artboardOrSliceForLayer: {
        value: function(sketchPage) {

            var layer = this.sketchLayer
            var isBlurred = this.hasStyle ? layer.style().blur().isEnabled() : false
            var exportLayer = layer

            if (isBlurred == false)
            {
                exportLayer = [layer duplicate]
                exportLayer.removeFromParent()
                sketchPage.addLayers([exportLayer])

                this.exportLayer = exportLayer
            }

            var absoluteInfluenceRect = exportLayer.absoluteInfluenceRect()
            var exportRequest = MSExportRequest.exportRequestsFromExportableLayer(exportLayer).firstObject()

            exportRequest.rect = absoluteInfluenceRect
            exportRequest.scale = this.exportInfo.exportOptions.imageExportScale
            exportRequest.includeArtboardBackground = true

            return exportRequest
        }
    },

    temporaryImagePath: {
        get: function () {
            var uuid = NSUUID.UUID().UUIDString()
            return Helpers.tempURL().URLByAppendingPathComponent(uuid).URLByAppendingPathExtension("png")
        }
    },

    contentsGravityForTextAlignment: {

        value: function(textAlignmentValue) {

            if (textAlignmentValue == 4) {
                return "left"
            }

            else if (textAlignmentValue == 2) {
                return "center"
            }

            else if (textAlignmentValue == 1) {
                return "right"
            }

            else {
                return "left"
            }
        }
    },

    firstActiveShadow: {
        get: function() {

            if (this.hasStyle == false) {
                return null
            }

            var shadows = this.sketchLayer.style().shadows()
            var activeShadow = null

            for (var i = 0; i < shadows.count(); i++)
            {
                var shadow = shadows.objectAtIndexOrNil(i)

                if (shadow && shadow.isEnabled()) {
                    activeShadow = shadow
                    break
                }
            }

            return activeShadow
        }
    },

    // -------------------------------------------------------------------------------------------
    // Archiving
    // -------------------------------------------------------------------------------------------

    archivableRepresentation: {
        value: function() {

            var arch = {}

            arch.uuid = this.uuid
            arch.name = this.name
            arch.className = this.className
            arch.hidden = this.hidden
            arch.locked = this.locked
            arch.bounds = Helpers.stringFromRect(this.bounds)
            arch.position = Helpers.stringFromPoint(this.position)
            arch.anchorPoint = Helpers.stringFromPoint(this.anchorPoint)
            arch.duration = this.duration
            arch.contentsGravity = this.contentsGravity
            arch.contentsScale = this.contentsScale
            arch.opacity = this.opacity
			arch.minificationFilter = this.minificationFilter
			arch.magnificationFilter = this.magnificationFilter

            // Shadows
            //
            if (this.exportsShadow && this.firstActiveShadow != null) {

                var shadow = this.firstActiveShadow
                var shadowColor = shadow.color();

                arch.shadowOpacity = 1.0 // No Sketch equivalent
                arch.shadowOffset = "{" + shadow.offsetX() + ", " + shadow.offsetY() + "}"
                arch.shadowRadius = shadow.blurRadius()

                if (shadowColor) {
                    arch.shadowColor = Helpers.archiveFromSketchColor(shadowColor)
                }
            }

            if (this.masksToBounds) {
                arch.masksToBounds = this.masksToBounds
            }

            if (this.contentIdentifier) {

                var contentImage = {
                    className: "JRContentImage",
                    contentIdentifier: this.contentIdentifier
                }

                arch.contentImage = contentImage
            }

            if (this.metadata) {
                arch.metadata = this.metadata
            }

            if (this.mask) {
                var archivedMask = this.mask.archivableRepresentation()

                if (archivedMask) {
                    arch.mask = archivedMask
                }
            }

            if (this.exportsSublayers)
            {
                var sublayers = []
                for (var sublayer of this.sublayers)
                {
                    // Mask layers are already archived via
                    // the 'mask' property and are not considered
                    // 'sublayers'
                    //
                    if (sublayer.masksParentLayer) {
                        continue
                    }

                    var archivedSublayer = sublayer.archivableRepresentation()

                    if (archivedSublayer) {
                        sublayers.push(archivedSublayer)
                    }
                }

                arch.sublayers = sublayers
            }

            return arch
        }
    }
})


// -------------------------------------------------------------------------------------------
// Parent Layer – Abstract parent class for Layers that have sublayers
// -------------------------------------------------------------------------------------------

ParentLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    Layer.call(this, sketchLayer, exportInfo)

    this.discoverSketchSublayers()
}

ParentLayer.prototype = Object.create(Layer.prototype, {

    discoverSketchSublayers: {
        value: function() {
            var sublayerArray = this.sketchLayer.layers()

            for (var layerIndex = 0; layerIndex < sublayerArray.count(); layerIndex++)
            {
            	var sketchSublayer = sublayerArray.objectAtIndex(layerIndex)
                var sublayer = this.layerForSketchLayer(sketchSublayer)

                print("Discovering sublayer for: [" + sketchSublayer + "] Kite sublayer: [" + sublayer + "]")

                if (sublayer != null) {
                    this.addSublayer(sublayer)
                }
            }
        },
    },

    addSublayer: {
        value: function(sublayer) {
            if (this.sublayers.indexOf(sublayer) == -1)
            {
                sublayer.parentLayer = this;
                this.sublayers.push(sublayer)
            }
        }
    },

    layerForSketchLayer: {
        value: function(sketchLayer) {

            if (!sketchLayer) { return null; }

            var layerClassString = sketchLayer.className()
            var exportInfo = this.exportInfo

            if (layerClassString == kTextLayerClass) {
                return new TextLayer(sketchLayer, exportInfo)
            }

            else if (layerClassString == kSliceLayerClass) {

                // Don't export Slice Layers
                //
                return null;
            }

            else if (layerClassString == kArtboardLayerClass) {
                return new ArtboardLayer(sketchLayer, exportInfo)
            }

            else if (layerClassString == kShapeGroupLayerClass) {
                return new ShapeGroupLayer(sketchLayer, exportInfo)
            }

            else if (layerClassString == kLayerGroupLayerClass) {
                return new LayerGroupLayer(sketchLayer, exportInfo)
            }

            else if (layerClassString == kBitmapLayerClass) {
                return new BitmapLayer(sketchLayer, exportInfo)
            }

            else if (layerClassString == kSymbolInstanceLayerClass) {
                return new SymbolInstanceLayer.fromDetachingLayer(sketchLayer, exportInfo)
            }

            else if (layerClassString == kSymbolMasterLayerClass) {
                return new SymbolMasterLayer(sketchLayer, exportInfo)
            }

            else if (layerClassString == kShapePathLayerClass) {
                return new ShapeGroupLayer(sketchLayer, exportInfo)
            }

            else if (layerClassString.match(/Shape$/)) {
                return new ShapeGroupLayer(sketchLayer, exportInfo)
            }

            else {
                log("Unknown layer type? " + sketchLayer.className().toString())
                return new Layer(sketchLayer, exportInfo)
            }
        },
    },

    hasSublayers: {
        get: function() {
            return this.sublayers.length > 0
        }
    },

    shouldFlattenSublayers: {
        get: function() {
            var layerNameHasFlattenSuffix = this.sketchLayer.name().trim().match(/[*][*]$/)
            return (this.exportInfo.exportOptions.flattenLayersWithSuffix && layerNameHasFlattenSuffix)
        }
    },
})


// -------------------------------------------------------------------------------------------
// Page Layer
// -------------------------------------------------------------------------------------------

PageLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    ParentLayer.call(this, sketchLayer, exportInfo)
}

PageLayer.prototype = Object.create(ParentLayer.prototype, {

    frame: {
        value: function() {
            // 'contentBounds()' returns a CGRect, so transform it
            // into a 'MSRect-like' object
            //
            var contentBounds = this.sketchLayer.contentBounds()
            var pageFrame = CGRectUnion(contentBounds, CGRectZero)

            return {
                x: function() { return pageFrame.origin.x },
                y: function() { return pageFrame.origin.y },
                width: function() { return pageFrame.size.width },
                height: function() { return pageFrame.size.height }
            };
        },
    },

    contentBounds: {
        get: function () {
            return this.sketchLayer.contentBounds()
        }
    },

    position: {
        get: function () {
            return {
                x: 0,
                y: 0
            }
        }
    },

    exportsContentImage: {
        get: function() {
            return false
        }
    },
})


// -------------------------------------------------------------------------------------------
// Artboard Layer
// -------------------------------------------------------------------------------------------

ArtboardLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    ParentLayer.call(this, sketchLayer, exportInfo)

    this.masksToBounds = true
    this.isRootDocumentLayer = false
}

ArtboardLayer.prototype = Object.create(ParentLayer.prototype, {

    artboardOrSliceForLayer: {
        value: function(sketchPage) {

            // Artboard is already in the correct format
            return this.sketchLayer
        }
    },

    exportsContentImage: {
        get: function() {
            return false
        }
    },

    // -------------------------------------------------------------------------------------------
    // Archiving
    // -------------------------------------------------------------------------------------------

    archivableRepresentation: {
        value: function() {

            var arch = ParentLayer.prototype.archivableRepresentation.call(this)

            if (this.sketchLayer.hasBackgroundColor())
            {
                arch.backgroundColor = Helpers.archiveFromSketchColor(this.sketchLayer.backgroundColor())
            }

            // If this artboard is acting as the root layer for the document, make
            // sure it has a (0,0) origin on the canvas when archived
            //
            if (this.isRootDocumentLayer) {
                arch.position = Helpers.stringFromPoint(CGPointZero)
            }

            return arch
        }
    },
})


// -------------------------------------------------------------------------------------------
// ShapeGroup Layer – Superclass is 'Layer' so we don't traverse sublayers
// -------------------------------------------------------------------------------------------
//
// If you draw a free-form Bezier path in Sketch or insert a pre-defined shape,
// this is the layer type you get
//

ShapeGroupLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    Layer.call(this, sketchLayer, exportInfo)

    this.path = this.discoverShapePath()
    this.fillRule = this.discoverWindingRule()
    this.miterLimit = this.hasStyle ? sketchLayer.style().miterLimit() : 0
    this.lineJoin = this.discoverLineJoinStyle()
    this.lineCap = this.discoverLineCapStyle()
    this.lineDashPattern = this.discoverLineDashPattern()
}

ShapeGroupLayer.prototype = Object.create(Layer.prototype, {

    layerClass: {
        get: function() {
            return "CAShapeLayer"
        }
    },

    exportsContentImage: {
        get: function() {
            return this.exportInfo.exportOptions.exportShapeLayersAsImages || this.firstActiveFillType == kFillTypeImage
        }
    },

    fillColor: {
        get: function() {
            var fill = this.firstActiveFill
            var fillColor = null

            if (fill && fill.fillType && fill.fillType() == kFillTypeSolidColor) {
                fillColor = fill.color()
            }

            if ( ! fillColor ) {
                fillColor = Helpers.sketchLikeColorWithRGBA(0, 0, 0, 0)
            }

            return fillColor
        }
    },

    strokeColor: {
        get: function() {
            var border = this.firstActiveBorder
            var strokeColor = null

            if (border) {
                strokeColor = border.color()
            }

            if ( ! strokeColor ) {
                strokeColor = Helpers.sketchLikeColorWithRGBA(0, 0, 0, 0)
            }

            return strokeColor
        }
    },

    lineWidth: {
        get: function() {
            var border = this.firstActiveBorder
            var lineWidth = 0

            if (border) {
                lineWidth = border.thickness()
            }

            return lineWidth
        }
    },

    // -------------------------------------------------------------------------------------------
    // Archiving
    // -------------------------------------------------------------------------------------------

    archivableRepresentation: {
        value: function() {

            var arch = Layer.prototype.archivableRepresentation.call(this)

            arch.fillColor = Helpers.archiveFromSketchColor(this.fillColor)
            arch.strokeColor = Helpers.archiveFromSketchColor(this.strokeColor)
            arch.lineWidth = this.lineWidth
            arch.lineJoin = this.lineJoin
            arch.lineCap = this.lineCap

            if (this.exportsContentImage == false)
            {
                var pathWriter = new ShapePathWriter(this)
                var pathString = pathWriter.svgPathString();

                if (pathString !== "") {
                    arch.path = pathString;
                }
            }

            // Non-zero is the default, so no need to
            // save it if it's not even-odd
            //
            if (this.fillRule != "non-zero") {
                arch.fillRule = this.fillRule
            }

            // 10 is the default for CAShapeLayers
            //
            if (this.miterLimit != kDefaultShapeLayerMiterLimit) {
                arch.miterLimit = this.miterLimit
            }

            if (this.lineDashPattern) {
                arch.lineDashPattern = this.lineDashPattern
            }

            return arch
        }
    },

    // -------------------------------------------------------------------------------------------
    // Helper
    // -------------------------------------------------------------------------------------------

    firstActiveFillType: {
        get: function() {

            var fill = this.firstActiveFill
            if (fill && fill.fillType) {
                return fill.fillType()
            }

            return -1
        }
    },

    firstActiveFill: {
        get: function() {
            var fills = this.hasStyle ? this.sketchLayer.style().fills() : []
            var activeFill = null

            for (var i = 0; i < fills.count(); i++)
            {
                var fill = fills.objectAtIndexOrNil(i)

                if (fill && fill.isEnabled()) {
                    activeFill = fill
                    break
                }
            }

            return activeFill
        }
    },

    firstActiveBorder: {
        get: function() {
            var borders = this.hasStyle ? this.sketchLayer.style().borders() : []
            var activeBorder = null

            for (var i = 0; i < borders.count(); i++)
            {
                var border = borders.objectAtIndexOrNil(i)

                if (border && border.isEnabled()) {
                    activeBorder = border
                    break
                }
            }

            return activeBorder
        }
    },

    discoverShapePath: {
        value: function() {

            var shapePath = null

            // Dupe layer before flattening
            //
            var sketchLayerCopy = this.sketchLayer.duplicate()

            if (sketchLayerCopy && sketchLayerCopy.decoratedBezierPathInBounds)
            {
                shapePath = sketchLayerCopy.decoratedBezierPathInBounds()
            }

            if (shapePath == null && sketchLayerCopy.pathInBounds && NSBezierPath.bezierPathWithPath)
            {
                shapePath = NSBezierPath.bezierPathWithPath(sketchLayerCopy.pathInBounds())
            }

            if (shapePath == null)
            {
                // Default to empty path
                shapePath =  NSBezierPath.bezierPath()
            }

            // Remove duped layer from the tree
            //
            sketchLayerCopy.removeFromParent()

            return shapePath
        }
    },

    discoverWindingRule: {
        value: function() {
            var windingRule = this.hasStyle ? this.sketchLayer.style().windingRule() : kNonZeroWindingRule;
            return (windingRule == kEvenOddWindingRule) ? "even-odd" : "non-zero"
        }
    },

    discoverLineJoinStyle: {
        value: function() {

            var joinStyle = this.hasStyle ? this.sketchLayer.style().borderOptions().lineJoinStyle() : 0

            switch(joinStyle)
            {
                case kCALayerLineJoinStyleMiter:
                    return "miter"
                    break

                case kCALayerLineJoinStyleRound:
                    return "round"
                    break

                case kCALayerLineJoinStyleBevel:
                    return "bevel"
                    break

                default:
                    return "miter"
            }
        }
    },

    discoverLineCapStyle: {
        value: function() {

            var capStyle = this.hasStyle ? this.sketchLayer.style().borderOptions().lineCapStyle() : 0

            switch(capStyle)
            {
                case kCALayerLineCapStyleButt:
                    return "butt"
                    break

                case kCALayerLineCapStyleRound:
                    return "round"
                    break

                case kCALayerLineCapStyleSquare:
                    return "square"
                    break

                default:
                    return "butt"
            }
        }
    },

    discoverLineDashPattern: {
        value: function() {

            if (this.hasStyle == false) { return null; }

            var borderOptions = this.sketchLayer.style().borderOptions()

            if (borderOptions.hasDashPattern() && borderOptions.dashPattern().length > 0) {
                return borderOptions.dashPattern()
            }

            return null
        }
    },
})


// -------------------------------------------------------------------------------------------
// LayerGroup Layer
// -------------------------------------------------------------------------------------------

LayerGroupLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    ParentLayer.call(this, sketchLayer, exportInfo)
}

LayerGroupLayer.prototype = Object.create(ParentLayer.prototype, {

    exportsContentImage: {
        get: function() {

            // Only export an image of ourselves if we are
            // flattening our sublayers
            //
            return (this.shouldFlattenSublayers) ? true : false
        }
    },

    exportsSublayers: {
        get: function() {
            // Only export our sublayers if we aren't flattening
            // our sublayers
            //
            return (this.shouldFlattenSublayers) ? false : true
        }
    },
})


// -------------------------------------------------------------------------------------------
// Shape Container Layer – Special Kite Layer
// -------------------------------------------------------------------------------------------

ShapeContainerLayer = function(canvasSize, exportInfo) {

    // Call super's constructor
    //
    // This layer isn't backed by a real Sketch layer
    //
    ParentLayer.call(this, new MockSketchLayer(canvasSize), exportInfo)

    this.className = "JRShapeContainerLayer"
    this.autoresizingMask = 18
    this.canvasInfo = null
}

ShapeContainerLayer.prototype = Object.create(ParentLayer.prototype, {

    exportsContentImage: {
        get: function() {
            return false
        }
    },
})


// -------------------------------------------------------------------------------------------
// Text Layer
// -------------------------------------------------------------------------------------------

TextLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    Layer.call(this, sketchLayer, exportInfo)

    this.string = sketchLayer.stringValue()
    this.fontSize = sketchLayer.fontSize()
    this.fontName = sketchLayer.fontPostscriptName()
    this.foregroundColor = sketchLayer.textColor()
    this.alignmentMode = this.textAlignmentModeForSketchTextAlignmentValue(sketchLayer.textAlignment())
    this.kerning = sketchLayer.kerning()
    this.automaticallyAdjustsContentsScale = true
	this.minificationFilter = kCAFilterNearest
	this.magnificationFilter = kCAFilterNearest
    this.wrapped = true

    // If we aren't exporting a content image, set the scale to the
    // current window's backing scale factor
    //
    if (this.exportsContentImage == false)
    {
        this.contentsScale = NSScreen.mainScreen().backingScaleFactor()
    }

    try {
        var paraStyle = sketchLayer.styleAttributes().objectForKey(NSParagraphStyleAttributeName)

        if (paraStyle) {
            this.paragraphStyle = new ParagraphStyle(paraStyle)
        }
    }
    catch(error) {
        delete this.paragraphStyle
    }
}

TextLayer.prototype = Object.create(Layer.prototype, {

    exportsContentImage: {
        get: function() {
            return this.exportInfo.exportOptions.exportTextLayersAsImages
        }
    },

    // If we are exporting an image of the layer, the shadow will be baked in the image,
    // so no need to add an additional native CALayer shadow
    //
    exportsShadow: {
        get: function() {
            return (this.exportInfo.exportOptions.exportTextLayersAsImages == false)
        }
    },

    layerClass: {
        get: function() {
            // When we only export an image of text, use CALayer because CATextLayer
            // has wierd rendering of its -contents property
            //
            return this.exportsContentImage ? "CALayer" : "CATextLayer"
        }
    },

    // -------------------------------------------------------------------------------------------
    // Archiving
    // -------------------------------------------------------------------------------------------

    exportedStringValue: {
        get: function() {
            if (this.kerning != 0) {
                return this.attributedStringExportValue
            }
            else {
                return this.stringExportValue
            }
        }
    },

    attributedStringExportValue: {
        get: function() {

            // Create line-by-line to use Cocoa key names
            //
            var firstAttributes = {}
            firstAttributes[NSFontAttributeName] = this.fontArchive
            firstAttributes[NSForegroundColorAttributeName] = Helpers.archiveFromSketchColor(this.foregroundColor)
            firstAttributes[NSKernAttributeName] = this.kerning

            // Archive our ParagraphStyle if we have one
            //
            if (this.paragraphStyle)
            {
                var archivedParaStyle = this.paragraphStyle.archivableRepresentation()
                firstAttributes[NSParagraphStyleAttributeName] = archivedParaStyle
            }

            var firstAttributesArchive = {
                range: "{0, " + this.string.length() + "}",
                attributes: firstAttributes
            }

            var attributes = [firstAttributesArchive]

            var value = {
                string: this.string,
                attributes: attributes,
            }

            var attrString = {
                type: "AttributedString",
                value: value
            }

            return attrString
        }
    },

    stringExportValue: {
        get: function() {
            return {
                type: "String",
                value: this.string,
            }
        }
    },

    fontArchive: {
        get: function() {
            return {
                fontName: this.fontName,
                pointSize: this.fontSize,
            }
        }
    },

    archivableRepresentation: {
        value: function() {

            var arch = Layer.prototype.archivableRepresentation.call(this)

            var textAlignment = this.sketchLayer.textAlignment()
            arch.contentsGravity = this.contentsGravityForTextAlignment(textAlignment)

            arch.alignmentMode = this.alignmentMode
            arch.foregroundColor = Helpers.archiveFromSketchColor(this.foregroundColor)
            arch.wrapped = this.wrapped
            arch.fontSize = this.fontSize
            arch.font = this.fontArchive

            if (this.exportsContentImage == false) {

                arch.automaticallyAdjustsContentsScale = this.automaticallyAdjustsContentsScale
                arch.contentsScale = this.contentsScale
                arch.string = this.exportedStringValue
            }
            else {
                arch.needsDisplayOnBoundsChange = false
            }

            return arch
        }
    },

    // -------------------------------------------------------------------------------------------
    // Helper
    // -------------------------------------------------------------------------------------------

    textAlignmentModeForSketchTextAlignmentValue: {

        value: function(textAlignmentValue) {

            if (textAlignmentValue == 4) {
                return "left"
            }

            else if (textAlignmentValue == 2) {
                return "center"
            }

            else if (textAlignmentValue == 1) {
                return "right"
            }

            else {
                return "left"
            }
        }
    },
})


// -------------------------------------------------------------------------------------------
// Shape Layer
// -------------------------------------------------------------------------------------------

ShapeLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    Layer.call(this, sketchLayer, exportInfo)
}

ShapeLayer.prototype = Object.create(Layer.prototype, {

})


// -------------------------------------------------------------------------------------------
// Bitmap Layer
// -------------------------------------------------------------------------------------------

BitmapLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    Layer.call(this, sketchLayer, exportInfo)
}

BitmapLayer.prototype = Object.create(Layer.prototype, {

})


// -------------------------------------------------------------------------------------------
// Slice Layer
// -------------------------------------------------------------------------------------------

SliceLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    Layer.call(this, sketchLayer, exportInfo)
}

SliceLayer.prototype = Object.create(Layer.prototype, {

    artboardOrSliceForLayer: {
        value: function(sketchPage) {

            // Slice is already in the correct format
            return this.sketchLayer
        }
    },
})


// -------------------------------------------------------------------------------------------
// Symbol Instance Layer
// -------------------------------------------------------------------------------------------

SymbolInstanceLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    LayerGroupLayer.call(this, sketchLayer, exportInfo)
}

SymbolInstanceLayer.prototype = Object.create(LayerGroupLayer.prototype, {

})

SymbolInstanceLayer.fromDetachingLayer = function(sketchLayer, exportInfo) {

    // De-symbolize this symbol layer to export it properly
    //
    var newGroupLayer = null;

    // Sketch 53 uses this new detaching method
    //
    if (sketchLayer.detachByReplacingWithGroupRecursively_withDocument_visitedSymbols)
    {
        newGroupLayer = sketch.fromNative(sketchLayer).detach({recursively:true}).sketchObject;
    }
    else if (sketchLayer.detachByReplacingWithGroupRecursively)
    {
        newGroupLayer = sketchLayer.detachByReplacingWithGroupRecursively()
    }

    if (newGroupLayer) {
        return new SymbolInstanceLayer(newGroupLayer, exportInfo)
    }
    else {
        return null
    }
}


// -------------------------------------------------------------------------------------------
// Symbol Master Layer
// -------------------------------------------------------------------------------------------

SymbolMasterLayer = function(sketchLayer, exportInfo) {

    // Call super's constructor
    //
    ArtboardLayer.call(this, sketchLayer, exportInfo)

    // We are like an ArtboardLayer that doesn't mask to bounds
    //
    this.masksToBounds = false
}

SymbolMasterLayer.prototype = Object.create(ArtboardLayer.prototype, {

})


// -------------------------------------------------------------------------------------------
// Mock Sketch Layer – Used by ShapeContainerLayer since it isn't backed by a real Sketch layer
// -------------------------------------------------------------------------------------------

MockSketchLayer = function(canvasSize) {

    this.isLocked = function () { return false }
    this.isVisible = function () { return true }
    this.name = function () { return "Background Layer" }
    this.hasClippingMask = function () { return false }
    this.frame = function() {
        return {
            x: function () { return 0 },
            y: function () { return 0 },
            width: function () { return canvasSize.width },
            height: function () { return canvasSize.height }
        }
    }
}

MockSketchLayer.prototype = Object.create(Object.prototype, {

    layers: {
        value: function() {
            return NSArray.array()
        },
    },
})
